"""This module defines the abstract class for the blueprint from all method's behaviours."""
from abc import ABC, abstractmethod


class EChemProcedure(ABC):
    """ Abstract class for the blueprint from all method's behaviours.

    Args:
        ABC (class): Abstract Base Class
    """
    @abstractmethod
    def analyze(self):
        """Abstract method for the analysis of the data.
        """

    @abstractmethod
    def plot(self, save_dir:str, plots:list, optional_name:str):
        """Abstract method for the plotting of the data.

        Args:
            save_dir (str): The directory where the plots are saved.
            plots (list): The plots that are saved.
            optional_name (str): The optional name of the plot.
        """

    @abstractmethod
    def save_data(self, save_dir:str, optional_name:str):
        """Abstract method for the saving of the data.

        Args:
            save_dir (str): The directory where the data is saved.
            optional_name (str): The optional name of the data.
        """

    @abstractmethod
    def perform_all_actions(self, save_dir:str, plots:list, optional_name:str):
        """Abstract method for the performing of all actions.

        Args:
            save_dir (str): The directory where the data is saved.
            plots (list): The plots that are saved.
            optional_name (str): The optional name of the data.
        """



"""This module defines the Arrhenius procedure"""
import os
import math
import numpy as np

from attrs import define, field
from attrs.setters import frozen
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error

from madap.logger import logger
from madap.utils import utils
from madap.echem.procedure import EChemProcedure
from madap.echem.arrhenius.arrhenius_plotting import ArrheniusPlotting as aplt


log = logger.get_logger("arrhenius")
@define
class Arrhenius(EChemProcedure):
    """ Class definition for visualization and analysis of Arrhenius equation.

    Attributes:
        temperatures (np.array): Array of temperatures in Celcius.
        conductivity (np.array): Array of conductivity in S/cm.
        gas_constant (float): Gas constant in [J/mol.K].
        activation (float): Activation energy in [mJ/mol].
        arrhenius_constant (float): Arrhenius constant in [S.cm⁻¹].
        inverted_scale_temperatures (np.array): Array of temperatures in 1000/K.
        fit_score (float): R2 score of the fit.
        ln_conductivity_fit (np.array): Array of log conductivity fit.
        intercept (float): Intercept of the fit.
        coefficients (float): Slope of the fit.
    """
    temperatures: list[float] = field(on_setattr=frozen)
    conductivity: list[float] = field(on_setattr=frozen)
    gas_constant = 8.314        # [J/mol.K]
    activation = None           # [mJ/mol]
    arrhenius_constant = None   # [S.cm⁻¹]
    inverted_scale_temperatures = None
    fit_score = None
    ln_conductivity_fit = None
    intercept = None
    coefficients = None
    mse_calc = None
    figure = None

    def analyze(self):
        """Analyze the data and fit the Arrhenius equation.
        """
        # the linear fit formula: ln(sigma) = -E/RT + ln(A)
        self._cel_to_thousand_over_kelvin()

        reg = LinearRegression().fit(self.inverted_scale_temperatures.values.reshape(-1,1), self._log_conductivity())
        self.fit_score = reg.score(self.inverted_scale_temperatures.values.reshape(-1,1), self._log_conductivity())
        self.coefficients, self.intercept = reg.coef_[0], reg.intercept_
        self.arrhenius_constant = math.exp(reg.intercept_)
        self.activation = reg.coef_[0]*(-self.gas_constant)
        self.ln_conductivity_fit = reg.predict(self.inverted_scale_temperatures.values.reshape(-1,1))
        self.mse_calc = mean_squared_error(self._log_conductivity(), self.ln_conductivity_fit)

        log.info(f"Arrhenius constant is {round(self.arrhenius_constant,4)} [S.cm⁻¹] \
                 and activation is {round(self.activation,4)} [mJ/mol] \
                 with the score {self.fit_score}")


    def plot(self, save_dir:str, plots:list, optional_name:str = None):
        """Plot the raw data and/or the results of the Arrhenius analysis.

        Args:
            save_dir (str): Directory where the plots should be saved.
            plots (list): List of plots included in the analysis.
            optional_name (str): Optional name for the analysis.
        """
        plot_dir = utils.create_dir(os.path.join(save_dir, "plots"))
        plot = aplt()

        fig, available_axes = plot.compose_arrhenius_subplot(plots=plots)
        for sub_ax, plot_name in zip(available_axes, plots):
            if plot_name == "arrhenius":
                plot.arrhenius(subplot_ax=sub_ax, temperatures= self.temperatures,
                               log_conductivity= self._log_conductivity(),
                               inverted_scale_temperatures = self.inverted_scale_temperatures)
            elif plot_name == "arrhenius_fit":
                plot.arrhenius_fit(subplot_ax = sub_ax, temperatures = self.temperatures, log_conductivity = self._log_conductivity(),
                                inverted_scale_temperatures = self.inverted_scale_temperatures,
                                #intercept = self.intercept, slope = self.coefficients,
                                ln_conductivity_fit=self.ln_conductivity_fit, activation= self.activation,
                                arrhenius_constant = self.arrhenius_constant, r2_score= self.fit_score)
            else:
                log.error("Arrhenius class does not have the selected plot.")
        fig.tight_layout()
        self.figure = fig
        name = utils.assemble_file_name(optional_name, self.__class__.__name__) if \
                    optional_name else utils.assemble_file_name(self.__class__.__name__)
        plot.save_plot(fig, plot_dir, name)

    def save_data(self, save_dir:str, optional_name:str = None):
        """Save the results of the analysis.

        Args:
            save_dir (str): Directory where the data should be saved.
            optional_name (str): Optional name for the analysis.
        """
        save_dir = utils.create_dir(os.path.join(save_dir, "data"))
        # Save the fitted circuit

        name = utils.assemble_file_name(optional_name, self.__class__.__name__, "linear_fit.json") if \
                optional_name else utils.assemble_file_name(self.__class__.__name__, "linear_fit.json")

        meta_data = {"R2_score": self.fit_score, "MSE": self.mse_calc, 'fit_slope': self.coefficients, "fit_intercept": self.intercept,
                    "arr_constant [S.cm⁻¹]": self.arrhenius_constant, "activation [mJ/mol]": self.activation,
                    "gas_constant [J/mol.K]": self.gas_constant}

        utils.save_data_as_json(directory=save_dir, name=name, data=meta_data)
        # Save the dataset
        data = utils.assemble_data_frame(**{"temperatures [\u00b0C]": self.temperatures,
                                            "conductivity [S/cm]": self.conductivity,
                                            "inverted_scale_temperatures [1000/K]": self.inverted_scale_temperatures,
                                            "log_conductivty [ln(S/cm)]": self._log_conductivity(),
                                            "log_conductivity_fit [ln(S/cm)]":self.ln_conductivity_fit})

        data_name = utils.assemble_file_name(optional_name, self.__class__.__name__, "data.csv") if \
                        optional_name else  utils.assemble_file_name(self.__class__.__name__, "data.csv")
        utils.save_data_as_csv(save_dir, data, data_name)


    def perform_all_actions(self, save_dir:str, plots:list, optional_name:str = None):
        """Wrapper function to perform all actions:\n
         - Analyze the data \n
         - Plot the data \n
         - Save the data

        Args:
            save_dir (str): Directory where the data should be saved.
            plots (list): plots to be included in the analysis.
            optional_name (str): Optional name for the analysis.
        """
        self.analyze()
        self.plot(save_dir=save_dir, plots=plots, optional_name=optional_name)
        self.save_data(save_dir=save_dir, optional_name=optional_name)

    @property
    def figure(self):
        """Get the figure of the analysis.

        Returns:
            obj: matplotlib.figure.Figure
        """
        return self._figure

    @figure.setter
    def figure(self, figure):
        """Setter for the figure attribute.

        Args:
            figure (obj): matplotlib.figure.Figure
        """
        self._figure = figure

    def _log_conductivity(self):
        """Convert the conductivity to log scale.

        Returns:
            np.array: Log of the conductivity.
        """
        return np.log(self.conductivity)

    def _cel_to_thousand_over_kelvin(self):
        """Convert the temperatures from Celcius to 1000/K.
        """
        converted_temps = 1000/(self.temperatures + 273.15)
        self.inverted_scale_temperatures = converted_temps



from madap.echem.voltammetry.voltammetry import Voltammetry
from madap.echem.procedure import EChemProcedure
from madap.logger import logger

log = logger.get_logger("cyclic_amperometry")


class Voltammetry_CA(Voltammetry, EChemProcedure):
    def __init__(self, voltage, current, time, capacity:list[float]=None, cycle=None) -> None:
        super().__init__(voltage, current, time)
        self.capacity = self._calculate_capacity() if capacity is None else capacity
        self.cycle = cycle
    def analyze(self):
        pass

    def plot(self):
        # TODO plots
        # TODO i vs t
        # TODO i vs Q
        # TODO currentDensity vs time
        # TODO Q vs cycle
        pass

    def save_data(self):
        pass

    def perform_all_actions(self, result_dir, plots):
        pass

    def _calculate_capacity(self):
        # TODO di/dt
        pass


"""Impedance Plotting module."""
import numpy as np
import matplotlib.colors as mcl
from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

from madap.logger import logger
from madap.plotting.plotting import Plots


log = logger.get_logger("impedance_plotting")

class ImpedancePlotting(Plots):
    """General Plotting class for Impedance method.

    Args:
        Plots (class): Parent class for plotting all methods.
    """
    def __init__(self) -> None:
        super().__init__()
        self.plot_type = "impedance"

    def nyquist(self, subplot_ax, frequency, real_impedance, imaginary_impedance,
                colorbar:bool=True, ax_sci_notation = None, scientific_limit=None,
                scientific_label_colorbar=False, legend_label=False,voltage:float = None,
                color_map:str="viridis", norm_color=None):
        """Defines the nyquist plot for raw data

        Args:
            subplot_ax (ax): Subplot axis
            frequency (np.array): Frequency array.
            real_impedance (np.array): Real impedance array.
            imaginary_impedance (np.array): Imaginary impedance array.
            colorbar (bool, optional): If True, adds a colorbar. Defaults to True.
            ax_sci_notation (bool, optional): If True, adds scientific notation to the axis. Defaults to None.
            scientific_limit (int, optional): If ax_sci_notation is True, defines the number of significant digits. Defaults to None.
            scientific_label_colorbar (bool, optional): If True, adds scientific notation to the colorbar. Defaults to False.
            legend_label (bool, optional): If True, adds a legend. Defaults to False.
            voltage (float, optional): Voltage of the circuit. Defaults to None.
            color_map (str, optional): Color map. Defaults to "viridis".
            norm_color (bool, optional): If True, normalizes the colorbar. Defaults to None.
        """

        log.info("Creating Nyquist plot")
        norm = mcl.LogNorm(vmin=min(frequency), vmax=max(frequency)) if norm_color else None

        label = f"v = {voltage} [V]" if (voltage and voltage != "")  else None
        nyquist_plot = subplot_ax.scatter(real_impedance, -imaginary_impedance,
                                          c=frequency, norm=norm, s=10,
                                          cmap=color_map, rasterized=True,
                                          label=label)

        nyquist_plot.set_clim(min(frequency), max(frequency))

        self.plot_identity(subplot_ax, xlabel=r"Z' $[\Omega]$", ylabel=r"-Z'' $[\Omega]$",
                            x_lim=[0, max(real_impedance)+200],
                            y_lim=[0, max(-imaginary_impedance)+200],
                            ax_sci_notation=ax_sci_notation,
                            scientific_limit=scientific_limit)

        if (legend_label and voltage is not None):
            _, labels = subplot_ax.get_legend_handles_labels()
            new_handles= [Line2D([0], [0], marker='o',
                        markerfacecolor="black",
                        markeredgecolor="black", markersize=3, ls='')]
            subplot_ax.legend(new_handles, labels, loc="upper left",fontsize=5.5)

        if colorbar:
            self.add_colorbar(nyquist_plot, subplot_ax, scientific_label_colorbar,
                            scientific_limit=scientific_limit,
                            colorbar_label=r"f $[Hz]$")

    def bode(self, subplot_ax, frequency, real_impedance, imaginary_impedance, phase_shift,
             ax_sci_notation=None, scientific_limit=None, log_scale='x'):
        """Defines the bode plot for raw data

        Args:
            subplot_ax (ax): Subplot axis
            frequency (np.array): Frequency array.
            real_impedance (np.array): Real impedance array.
            imaginary_impedance (np.array): Imaginary impedance array.
            phase_shift (np.array): Phase shift array.
            ax_sci_notation (bool, optional): If True, adds scientific notation to the axis. Defaults to None.
            scientific_limit (int, optional): If ax_sci_notation is True, defines the number of significant digits. Defaults to None.
            log_scale (str, optional): If 'x', plots the x axis in log scale. Defaults to 'x'.
        """
        log.info("Creating Bode plot")
        impedance_magnitude = np.sqrt(real_impedance**2 +  imaginary_impedance**2)
        subplot_ax.scatter(frequency, impedance_magnitude, rasterized=True, s=10,c="#453781ff")
        subplot_ax.tick_params(axis="y", colors="#453781ff")
        ax2 = subplot_ax.twinx()
        ax2.scatter(frequency, -phase_shift, rasterized=True, s=10, c="#20a387ff")
        ax2.tick_params(axis="y", colors="#20a387ff")
        self.plot_identity(subplot_ax, xlabel=r"f $[Hz]$", ylabel=r"|Z| $[\Omega]$",
                        ax_sci_notation=ax_sci_notation,
                        scientific_limit=scientific_limit,
                        log_scale=log_scale)
        self.plot_identity(ax2, ylabel="\u03c6 [\u00b0]",
                        y_lim=[self.round_tenth(-phase_shift)[0],
                               self.round_tenth(-phase_shift)[1]],
                        log_scale=log_scale, step_size_y=10)

    def nyquist_fit(self, subplot_ax, frequency, real_impedance, imaginary_impedance,
                    fitted_impedance, chi, suggested_circuit, colorbar:bool=True,
                    ax_sci_notation = None, scientific_limit:int=3,
                    scientific_label_colorbar=False, legend_label=False,
                    voltage:float = None, color_map:str="viridis",
                    norm_color=None):
        """Defines the nyquist plot for fitted data

        Args:
            subplot_ax (ax): Subplot axis
            frequency (np.array): Frequency array.
            real_impedance (np.array): Real impedance array.
            imaginary_impedance (np.array): Imaginary impedance array.
            fitted_impedance (np.array): Fitted impedance array.
            chi (float): Chi value of the fit.
            suggested_circuit (str): The string definition of the suggested circuit.
            colorbar (bool, optional): If True, adds a colorbar. Defaults to True.
            ax_sci_notation (bool, optional): If True, adds scientific notation to the axis. Defaults to None.
            scientific_limit (int, optional): If ax_sci_notation is True, defines the number of significant digits. Defaults to None.
            scientific_label_colorbar (bool, optional): If True, adds scientific notation to the colorbar. Defaults to False.
            legend_label (bool, optional): If True, adds a legend. Defaults to False.
            voltage (float, optional): Voltage of the circuit. Defaults to None.
            color_map (str, optional): Color map. Defaults to "viridis".
            norm_color (bool, optional): If True, normalizes the colorbar. Defaults to None.
        """
        log.info("Creating a fitted Nyquist plot")
        nyquist_label = fr" v = {voltage} [V], \
                          $\chi_{{linKK}}^{2}$ = {np.format_float_scientific(chi, 3)}" \
                          if voltage else fr"$\chi^{2}$ = {np.format_float_scientific(chi, 3)}"

        norm = mcl.LogNorm(vmin=min(frequency), vmax=max(frequency)) if norm_color else None
        nyquist_plot = subplot_ax.scatter(real_impedance, -imaginary_impedance,
                                        c=frequency, norm=norm, s=10,
                                        cmap=color_map,rasterized=True,
                                        label=nyquist_label)

        subplot_ax.plot(np.real(fitted_impedance), -np.imag(fitted_impedance),
                        label=f"Fitted with {suggested_circuit}", color="k")

        self.plot_identity(subplot_ax, xlabel=r"Z' $[\Omega]$", ylabel=r"-Z'' $[\Omega]$",
                               x_lim=[0, max(real_impedance) +200],
                               y_lim=[0, max(-imaginary_impedance) +200],
                               ax_sci_notation=ax_sci_notation,
                               scientific_limit=scientific_limit)

        if legend_label:
            _, labels = subplot_ax.get_legend_handles_labels()
            new_handles= [Line2D([0], [0], marker='o',
                          markerfacecolor="black",
                          markeredgecolor="black", markersize=3, ls=''),
                          Line2D([0], [0],
                          color="black", markersize=3, lw=1)]
            subplot_ax.legend(new_handles, labels, loc="upper left", fontsize=5.5)

        if colorbar:
            self.add_colorbar(nyquist_plot, subplot_ax, scientific_label_colorbar,
                              scientific_limit=scientific_limit,
                              colorbar_label=r"f $[Hz]$")
        self.ax = subplot_ax

    def residual(self, subplot_ax, frequency, res_real, res_imag,
                log_scale='x', ax_sci_notation = None,
                scientific_limit:int=3):
        """Defines the residual plot for raw data

        Args:
            subplot_ax (ax): Subplot axis.
            frequency (np.array): Frequency array.
            res_real (np.array): Real residual array.
            res_imag (np.array): Imaginary residual array.
            log_scale (str, optional): If 'x', plots the x axis in log scale. Defaults to 'x'.
            ax_sci_notation (bool, optional): If True, adds scientific notation to the axis. Defaults to None.
            scientific_limit (int, optional): If ax_sci_notation is True, defines the number of significant digits. Defaults to None.
        """

        log.info("Creating a residual plot")
        subplot_ax.plot(frequency, res_real, label=r"$\Delta_{real}$", color="#453781ff",
                linestyle="--", marker='o')
        subplot_ax.plot(frequency, res_imag, label=r"$\Delta_{imaginary}$", color="#20a387ff",
                linestyle="--", marker='o')

        self.plot_identity(subplot_ax, xlabel=r"f $[Hz]$", ylabel=r"$\Delta$(residuals)",
                                ax_sci_notation=ax_sci_notation,
                                scientific_limit=scientific_limit, log_scale=log_scale)
        subplot_ax.legend(loc="lower right", fontsize=5.5)

    def compose_eis_subplot(self, plots:list):
        """Compose the EIS subplot

        Args:
            plots (list): List of plots to be composed.

        Returns:
            fig, ax: Figure and axis of the subplot.
        """

        plt.close('all')
        if len(plots)==1:
            fig = plt.figure(figsize=(3.5,3))
            spec = fig.add_gridspec(1, 1)
            ax = fig.add_subplot(spec[0,0])
            return fig, [ax]

        if len(plots) == 2:
            fig_size = 9 if ("nyquist" and "nyquist_fit") in plots else 8.5
            fig = plt.figure(figsize=(fig_size, 4))
            spec = fig.add_gridspec(1, 2)
            ax1 = fig.add_subplot(spec[0, 0])
            ax2= fig.add_subplot(spec[0, 1])
            return fig, [ax1, ax2]

        if len(plots) == 3:
            fig_size= 7 if ("nyquist" and "nyquist_fit" and "bode") in plots else 6.5
            fig = plt.figure(figsize=(fig_size, 5))
            spec = fig.add_gridspec(2, 2)
            if "residual" in plots:
                ax1 = fig.add_subplot(spec[0, 0])
                ax2 = fig.add_subplot(spec[0, 1])
                ax3 = fig.add_subplot(spec[1, :])
            else:
                ax1 = fig.add_subplot(spec[0, 0])
                ax2 = fig.add_subplot(spec[1, 0])
                ax3 = fig.add_subplot(spec[:, 1])
            return fig, [ax1, ax2, ax3]

        if len(plots) == 4:
            fig = plt.figure(figsize=(7.5, 6))
            spec = fig.add_gridspec(2, 2)
            ax1 = fig.add_subplot(spec[0, 0])
            ax2= fig.add_subplot(spec[0, 1])
            ax3 = fig.add_subplot(spec[1, 0])
            ax4 = fig.add_subplot(spec[1, 1])
            return fig, [ax1, ax2, ax3, ax4]

        if len(plots) == 0:
            log.error("No plots for EIS were selected.")
            return Exception(f"No plots for EIS were selected for plot {self.plot_type}.")

        log.error("Maximum plots for EIS is exceeded.")
        return Exception(f"Maximum plots for EIS is exceeded for plot {self.plot_type}.")



""" This module is the main entry point for MADAP. It defines the CLI to be used by the user."""
import os
import sys
import argparse
import re

from pathlib import Path
import pandas as pd

from madap.utils import utils
from madap.logger import logger
from madap.data_acquisition import data_acquisition as da
from madap.echem.e_impedance import e_impedance
from madap.echem.arrhenius import arrhenius
from madap.echem.voltammetry import voltammetry_CA



log = logger.setup_applevel_logger(file_name = 'madap_debug.log')

def _analyze_parser_args():
    """Private function to analyze the parser arguments

    Returns:
        argparse: The parser with correct arguments
    """
    first_parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    # Options for procedure
    procedure = first_parser.add_argument_group("Options for the procedure")
    procedure.add_argument("-p", "--procedure", type=str, choices=['arrhenius', 'impedance', 'voltammetry'],
                          help="Procedure of the analysis")
    proc = first_parser.parse_known_args()[0]
    if proc.procedure == "impedance":
        procedure.add_argument("-ip", "--impedance_procedure", type=str, required=True, choices=['EIS', 'Mottschotcky', 'Lissajous'],
                            help="Which of the impedance procedures you want to use?")
        proc = first_parser.parse_known_args()[0]
        if proc.impedance_procedure == "EIS":
            eis = first_parser.add_argument_group("Options for the EIS procedure")
            # Add the arguments for the EIS procedure
            eis.add_argument("-pl", "--plots", required=True, choices=["nyquist" ,"nyquist_fit", "residual", "bode"],
                            nargs="+", help="plots to be generated")
            eis.add_argument("-v", "--voltage", type=float, required=False, default=None,
                            help="applied voltage [V] if applicable")
            eis.add_argument("-cc", "--cell_constant", type=float, required=False, default=None,
                            help="cell constant if applicable")
            eis.add_argument("-sc", "--suggested_circuit", type=str, required=False, default=None,
                            help="suggested circuit if applicable. \n Available elements are 's', 'C', 'Ws', 'K', 'W', 'Wo', 'R', \
                                'p', 'L', 'TLMQ', 'CPE', 'G', 'La', 'T', 'Gs' \n Parallel circuit can be defined like p(element1, element2)\
                                and the series circuit like element1_element2")
            eis.add_argument("-iv", "--initial_values", required=False, default=None,
                            help="initial values for the suggested circuit. \
                                \n format: element1, element2, ... \
                                \n it will be used just if the suggested_circuit is available.")

        elif proc.impedance_procedure == "Mottschotcky":
            # TODO
            pass
        elif proc.impedance_procedure == "Lissajous":
            # TODO
            pass

    elif proc.procedure == "arrhenius":
        arrhenius_pars = first_parser.add_argument_group("Options for the Arrhenius procedure")
        arrhenius_pars.add_argument("-pl", "--plots", choices=["arrhenius" ,"arrhenius_fit"],
                                nargs="+", required=True, help="Plots to be generated")

    elif proc.procedure == "voltammetry":
        voltammetry_pars = first_parser.add_argument_group("Options for the Arrhenius procedure")
        voltammetry_pars.add_argument("-vp", "--voltammetry_procedure", type=str, required=True,
                            choices=['cyclic_voltammetric', 'cyclic_amperometric', "cyclic_potentiometric"],)
        proc = first_parser.parse_known_args()[0]
        if proc.voltammetry_procedure == "cyclic_voltammetric":
            # TODO
            pass
        elif proc.voltammetry_procedure == "cyclic_amperometric":
            # TODO
            pass
        elif proc.voltammetry_procedure == "cyclic_potentiometric":
            # TODO
            pass

    # Options for data import
    data = first_parser.add_argument_group("Options for data import")
    data.add_argument("-f", "--file", type=Path, required=True, metavar="FILE", help="Path to the data file")
    data.add_argument("-u", "--upper_limit_quantile", type=float, required=False,
                      default=0.99, help="Upper quantile for detecting the outliers in data")
    data.add_argument("-l", "--lower_limit_quantile", type=float, required=False,
                      default=0.01, help="Lower quantile for detecting the outliers in data")
    data_selection = data.add_mutually_exclusive_group()
    data_selection.add_argument("-sp", "--specific", type=str, nargs="+",
                            help="row and column number of the frequency, real impedance, \
                            imaginary_impedance and phase shift \
                            \n write n if it is not applicable \n order is important.\
                            \n format: start_row,end_row,start_column,end_column \
                            \n 1,10,1,2 means rows 1 to 10 and columns 1 to 2")
    data_selection.add_argument("-hl", "--header_list", type=str, nargs="+",
                            help="Definitions of the headers for frequency [Hz], \
                            real impedance [\u2126],\
                            imaginary_impedance [\u2126] and phase shift \u03c6 [\u00b0] \
                            \n write n if it is not applicable \n order is important.")

    parser = argparse.ArgumentParser(description='Use MADAP for electrochemical analysis',
                                     parents=[first_parser],
                                    formatter_class=argparse.RawDescriptionHelpFormatter)
    # Options for results
    parser.add_argument("-r", "--results", type=Path, required=True,
                        help="Directory for saving results")

    return parser


def call_impedance(data, result_dir, args):
    """calling the impedance procedure and parse the corresponding arguments

    Args:
        data (class): the given data frame for analysiswrite
        result_dir (str): the directory for saving results
        args (parser.args): Parsed arguments
    """

    if args.header_list:
        # Check if args header is a list
        if isinstance(args.header_list, list):
            header_names = args.header_list[0].split(", ") if len(args.header_list) == 1 else \
            args.header_list
        else:
            header_names = args.header_list

        phase_shift_data = None if len(header_names) == 3 else data[header_names[3]]

        _, nan_indices = da.remove_outlier_specifying_quantile(df = data,
                                                           columns = [header_names[1],
                                                                       header_names[2]],
                                                           low_quantile = args.lower_limit_quantile,
                                                           high_quantile = args.upper_limit_quantile)
        # remove nan rows
        data = da.remove_nan_rows(data, nan_indices)
        # extracting the data
        freq_data, real_data, imag_data = data[header_names[0]],\
                                          data[header_names[1]],\
                                          data[header_names[2]]

    if args.specific:

        try:
            if len(args.specific) >= 3:
                row_col = args.specific
            else:
                row_col = re.split('; |;', args.specific[0])

        except ValueError as e:
            log.error("The format of the specific data is not correct. Please check the help.")
            raise e

        selected_data = data.iloc[int(row_col[0].split(',')[0]): int(row_col[0].split(',')[1]), :]


        phase_shift_data = None if len(row_col) == 3 else da.select_data(data, row_col[3])


        freq_data, real_data, imag_data = da.select_data(selected_data, row_col[0]), \
                                          da.select_data(selected_data, row_col[1]), \
                                          da.select_data(selected_data, row_col[2])

        unprocessed_data = pd.DataFrame({"freq": freq_data, "real": real_data, "imag": imag_data})

        _, nan_indices = da.remove_outlier_specifying_quantile(df = unprocessed_data,
                                            columns = ["real", "imag"],
                                            low_quantile = args.lower_limit_quantile,
                                            high_quantile = args.upper_limit_quantile)

        data = da.remove_nan_rows(unprocessed_data, nan_indices)
        freq_data, real_data, imag_data = data["freq"], data["real"], data["imag"]

    impedance = e_impedance.EImpedance(da.format_data(freq_data), da.format_data(real_data),
                                da.format_data(imag_data), da.format_data(phase_shift_data))

    if args.impedance_procedure == "EIS":
        log.info(f"The given voltage is {args.voltage} [V], cell constant is {args.cell_constant},\
                   suggested circuit is {args.suggested_circuit} \
                   and initial values are {args.initial_values}.")

        # Instantiate the procedure
        procedure = e_impedance.EIS(impedance, voltage=args.voltage,
                                    suggested_circuit=args.suggested_circuit,
                                    initial_value=eval(args.initial_values)
                                    if args.initial_values else None,
                                    cell_constant=args.cell_constant)

    elif args.impedance_procedure == "Mottschotcky":
        #TODO
        # # Instantiate the procedure
        pass
    elif args.impedance_procedure == "Lissajous":
        #TODO
        # # Instantiate the procedure
        pass

    # Format plots arguments
    plots = da.format_plots(args.plots)

    # Perform all actions
    procedure.perform_all_actions(result_dir, plots=plots)

    return procedure

def call_arrhenius(data, result_dir, args):
    """Calling the arrhenius procedure and parse the corresponding arguments

    Args:
        data (class): the given data frame for analysis
        result_dir (str): the directory for saving results
        args (parser.args): Parsed arguments
    """


    if args.header_list:
        if isinstance(args.header_list, list):
            header_names = args.header_list[0].split(", ") if len(args.header_list) == 1 else \
            args.header_list
        else:
            header_names = args.header_list

        temp_data, cond_data = data[header_names[0]], data[header_names[1]]
    if args.specific:

        try:
            if len(args.specific) == 2:
                row_col = args.specific
            else:
                row_col = re.split('; |;', args.specific[0])

        except ValueError as e:
            log.error("The format of the specific data is not correct. Please check the help.")
            raise e

        selected_data = data.iloc[int(row_col[0].split(',')[0]): int(row_col[0].split(',')[1]), :]

        #row_col = args.specific[0].split(", ")

        temp_data, cond_data = da.select_data(selected_data, row_col[0]), \
                               da.select_data(selected_data, row_col[1])
        if (not isinstance(temp_data, pd.Series)) and (not isinstance(cond_data, pd.Series)):
            temp_data, cond_data = pd.Series(temp_data).astype(float), pd.Series(cond_data).astype(float)
    # Instantiate the procedure
    arrhenius_cls = arrhenius.Arrhenius(da.format_data(temp_data), da.format_data(cond_data))

    # Format the plots arguments
    plots = da.format_plots(args.plots)

    # Perform all actions
    arrhenius_cls.perform_all_actions(result_dir, plots = plots)

    return arrhenius_cls

def call_voltammetry(data, result_dir, plots):
    """ Calling the voltammetry procedure and parse the corresponding arguments

    Args:
        data (class): the given data frame for analysis
        result_dir (str): the directory for saving results
        plots (list): list of plots to be generated
    """
    log.info("What is the name (or index) of the column of voltage (v [V]) ?")
    # TODO
    voltage_idx = "voltage" #input()
    # TODO
    log.info("What is the name (or index) of the column of current (I [A]) ?")
    current_idx = "current" #input()
    # TODO
    log.info("What is the name (or index) of the column of time (t [s]) ?")
    time_idx = "time" #input()

    voltammetry_cls = voltammetry_CA.Voltammetry_CA(da.format_data(data[voltage_idx]),
                                  da.format_data(data[current_idx]),
                                da.format_data(data[time_idx]))
    if isinstance(plots, str):
        plots = [plots]
    if isinstance(plots, tuple):
        plots = list(plots)
    voltammetry_cls.perform_all_actions(result_dir, plots=plots)

def start_procedure(args):
    """Function to prepare the data for analysis.
    It also prepares folder for results and plots.

    Args:
        args (object): Object containing arguments from parser or gui.
    """

    data = da.acquire_data(args.file)
    da.remove_unnamed_col(data)
    log.info(f"the header of your data is: \n {data.head()}")
    result_dir = utils.create_dir(os.path.join(args.results, args.procedure))

    if args.procedure in ["impedance", "Impedance"]:
        procedure = call_impedance(data, result_dir, args)

    elif args.procedure in ["arrhenius", "Arrhenius"]:
        procedure = call_arrhenius(data, result_dir, args)

    elif args.procedure == "voltammetry":
        log.info("Voltammetrys is not supported at the moment. Exiting ...")
        sys.exit()

    return procedure

def main():
    """Main function to start the program.
    """
    log.info("==================================WELCOME TO MADAP==================================")
    # Create the parser
    parser = _analyze_parser_args()
    # Parse the argument
    args = parser.parse_args()

    # Acquire data
    start_procedure(args)


if __name__ == "__main__":
    main()



import os
from attrs import define, field
from attrs.setters import frozen
from madap.logger import logger
from madap.echem.procedure import EChemProcedure


log = logger.get_logger("voltammetry")

@define
class Voltammetry(EChemProcedure):
    voltage = field(default=0)
    current = field(default=0)
    time = field(default=0)

    def _calculate_capacity(self):
        # TODO di/dt
        pass

    def _calculate_columbic_efficiency(self):
        # TODO
        pass


""" This module handels the general plotting functions for the MADAP application. """
import os
import math

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

from madap.logger import logger


mpl.use('svg')
log = logger.get_logger("plotting")
class Plots():
    """_General class for multipurpose plotting
    """
    def __init__(self) -> None:

        mpl.rcParams.update(mpl.rcParamsDefault)
        mpl.rc('font', size=20)
        mpl.rc('axes', titlesize=20)
        style_path, _ =os.path.split(__file__)
        plt.style.use([os.path.join(style_path, 'styles', 'nature.mplstyle'),
                       os.path.join(style_path, 'styles', 'science.mplstyle'),
                       os.path.join(style_path, 'styles', 'no-latex.mplstyle')])
        plt.rcParams['text.usetex'] = False
        plt.rcParams['xtick.direction'] = 'in'
        plt.rcParams['ytick.direction'] = 'in'
        self.plot_type = ""
        self.ax = None

    def plot_identity(self, ax, xlabel:str=None, ylabel:str=None, x_lim:list=None, y_lim:list=None,
                      rotation:float=0, ax_sci_notation:bool=False, scientific_limit=0,
                      log_scale:str=None, step_size_x="auto", step_size_y="auto"):
        """Defines the "identity" of the plot.
        This includes the x and y labels, the x and y limits, the rotation of the x labels,
        wether or not scientific notation should be used,
        the log scale and the step size of the x and y axis.

        Args:
            ax (matplotlib.axes): axis to which the identity line should be added
            xlabel (str, optional): label of the x-axis. Defaults to None.
            ylabel (str, optional): label of the y-axis. Defaults to None.
            x_lim (list, optional): limits of the x-axis. Defaults to None.
            y_lim (list, optional): limits of the y-axis. Defaults to None.
            rotation (float, optional): rotation of the x and ylabels. Defaults to 0.
            ax_sci_notation (bool, optional): whether or not scientific notation should be used.
            scientific_limit (int, optional): scientific notation limit. Defaults to 0.
            log_scale (str, optional): log scale of the x and y axis. Defaults to None.
            Can be "x", "y"
            base10 (bool, optional): base 10 log scale. Defaults to False.
            step_size_x (str, optional): step size of the x axis. Defaults to "auto".
            step_size_y (str, optional): step size of the y axis. Defaults to "auto".
        """
        if xlabel:
            ax.set_xlabel(xlabel, fontsize=9)
        if ylabel:
            ax.set_ylabel(ylabel, fontsize=9)
        if x_lim:
            step = (max(x_lim)-min(x_lim))/5 if step_size_x=="auto" else step_size_x
            ax.set_xlim([min(x_lim), max(x_lim)])
            ax.set_xticks(np.arange(min(x_lim), max(x_lim), step))
        if y_lim:
            step = (max(y_lim)-min(y_lim))/5 if step_size_y=="auto" else step_size_y
            ax.set_ylim([min(y_lim), max(y_lim)])
            ax.set_yticks(np.arange(min(y_lim), max(y_lim), step))
        if rotation:
            ax.xaxis.set_tick_params(rotation=rotation)
        if ax_sci_notation:
            ax.ticklabel_format(style='sci', axis=ax_sci_notation, \
                                scilimits=(scientific_limit, scientific_limit))
        if log_scale:
            ax.set_xscale('log') if log_scale=='x' else (ax.set_yscale('log') if log_scale=='y' else (ax.set_xscale('log'), ax.set_yscale('log')))


    def add_colorbar(self, plot, ax, scientific_label_colorbar=None, scientific_limit=3, colorbar_label=None):
        """Adds a colorbar to a plot

        Args:
            plot (matplotlib.pyplot): plot to which the colorbar should be added
            ax (matplotlib.axes): axis to which the colorbar should be added
            scientific_label_colorbar (str, optional): whether or not the colorbar should be written
            scientific_limit (int, optional): scientific notation limit. Defaults to 3.
            colorbar_label (str, optional): label of the colorbar. Defaults to None.
        """

        color_bar = plt.colorbar(plot, ax=ax)

        if colorbar_label:
            color_bar.set_label(colorbar_label, fontsize=7)
            color_bar.ax.yaxis.set_label_position('right')


        if scientific_label_colorbar:
            color_bar.formatter.set_powerlimits((scientific_limit, scientific_limit))
            color_bar.update_ticks()

    def round_hundredth(self, nums):
        """Rounds a number to the nearest hundredth

        Args:
            nums (float): number to be rounded

        Returns:
            float: rounded number
        """
        if not nums.all() < 0:
            min_num, max_num = int(math.ceil(min(nums) / 100.0)) * 100 - 100, \
                int(math.ceil(max(nums) / 100.0)) * 100
        else:
            min_num, max_num = int(math.ceil(min(nums) / 100.0)) * 100 , \
                int(math.ceil(max(nums) / 100.0)) * 100+ 100
        return min_num, max_num

    def round_tenth(self, nums):
        """Rounds a number to the nearest tenth

        Args:
            nums (float): number to be rounded

        Returns:
            float: rounded number
        """
        if not nums.all() < 0:
            min_num, max_num = round(min(nums), -1) - 10, round(max(nums), -1)
        else:
            min_num, max_num = round(min(nums), -1) , round(max(nums), -1) + 10
        return min_num, max_num

    def set_xtick_for_two_axes(self, ax1, ax2, ax1_ticks, ax2_ticks, invert_axes=False):
        """Sets the xticks for two axes

        Args:
            ax1 (matplotlib.axes): first axis
            ax2 (matplotlib.axes): second axis
            ax1_ticks (list): ticks for the first axis
            ax2_ticks (list): ticks for the second axis
            invert_axes (bool, optional): whether or not the axes should be inverted. Default False.
        """
        ax1.set_xlim(ax2.get_xlim())
        ax1.set_xticks(ax2_ticks)
        ax1.set_xticklabels(ax1_ticks)
        if invert_axes:
            ax1.invert_xaxis()
            ax2.invert_xaxis()

    def save_plot(self, fig, directory, name):
        """Saves a plot

        Args:
            fig (matplotlib.pyplot): figure to be saved
            directory (str): directory in which the plot should be saved
            name (str): name of the plot
        """
        log.info(f"Saving .png and .svg in {directory}")
        fig.savefig(os.path.join(directory, f"{name}.svg"), dpi=900)
        fig.savefig(os.path.join(directory, f"{name}.png"), dpi=900)



""" Just a file containing some lengthy text used for the GUI helpers. """

HEADER_OR_SPECIFIC_HELP="If the data selection is 'Headers', insert the headers for: " \
                        "\nfrequency [Hz], real impedance [\u2126], imaginary impedance, phase shift \u03c6[\u00b0] (optional).Order is relevant.\n"\
                        "Example: 'freq, real, imag'" \
                        "\n \n"\
                        "If the data selection is 'Specific' insert the row and column numbers. Order is relevant." \
                        "\nFormat: 'start_row,end_row,start_column,end_column': 1,10,1,2  translates to rows 1 to 10 and columns 1 to 2."\
                        "\nExample: '0,40,0,1; 0,40,1,2; 0,40,2,3'"

SUGGESTED_CIRCUIT_HELP="Suggested circuit (optional).\n" \
                        "Available circuit elements are: 's', 'C', 'Ws', 'K', 'W', 'Wo', 'R'," \
                        "'p', 'L', 'TLMQ', 'CPE', 'G', 'La', 'T', 'Gs'." \
                        "\nParallel circuit can be defined as p(element1, element2) " \
                        "and the series circuit like element1_element2."\
                        "\nFormat: R0-p(R1,CPE1)"

INITIAL_VALUES_HELP="Initial values for the suggested circuit." \
                    "\nIt will be used just if the suggested circuit was defined. " \
                    "\nFormat: [element1, element2, ...], i.e.: [800,1e+14,1e-9,0.8] "


VOLTAGE_HELP="Applied voltage (optional)"

CELL_CONSTANT_HELP="Cell constant (optional)"

UPPER_LIMIT_QUANTILE_HELP="Upper limit quantile (optional)."

LOWER_LIMIT_QUANTILE_HELP="Lower limit quantile (optional)."

def plotting_element(subparser, plots):
    """ Adds an element to the GUI for plotting the data.

    Args:
        subparser (GooeyParser): Subparser to add the element to.
        plots (list): List of plot types.
    """
    subparser.add_argument("-pl", "--plots", metavar="Plots", required=True, choices=plots,
                             nargs="+", help="Plots to be generated", widget="Listbox")

def data_import_element(subparser):
    """ Adds an element to the GUI for importing the data.

    Args:
        subparser (GooeyParser): Subparser to add the element to.
    """
    # Procedure subgroup
    data = subparser.add_argument_group("Data import",
                                        gooey_options={'show_border': True, 'columns': 1})

    # Procedure - File
    file = data.add_argument_group("File selection", gooey_options={'show_border': True})
    file.add_argument("-f", "--file", type=str, required=True,
                      metavar="File containing the dataset", help="Path to the data file",
                      widget="FileChooser")

    # Procedure - File - Data Selection
    file_data = data.add_argument_group("Data Selection (select one)", gooey_options={
         'show_border': True,
         'columns': 1
     })
    file_data.add_argument("-ds", "--data-selection", choices=["Headers", "Specific"],
                           required=True, metavar="Data Selection",
                           help="Select all data (by header) or specific data", widget="Dropdown")
    file_data.add_argument("-dl", "--data-list", type=str, required=False,
                           metavar="Headers or Specific", help=HEADER_OR_SPECIFIC_HELP)



from madap.echem.voltammetry.voltammetry import Voltammetry
from madap.echem.procedure import EChemProcedure
from madap.logger import logger

log = logger.get_logger("cyclic_voltammetry")


class Voltammetry_CV(Voltammetry, EChemProcedure):
    def __init__(self, voltage, current, time) -> None:
        super().__init__(voltage, current, time)

    def analyze():
        # TODO
        pass

    def plot():
        # TODO plots
        # TODO V vs I
        pass

    def save_data():
        pass

    def perform_all_actions():
        pass
    def _peek_to_peek_seperation(self):
        pass
    def _calcualte_diffusion_coefficient(self):
        pass
    def _randles_sevcik_equation(self):
        pass


from madap.echem.voltammetry.voltammetry import Voltammetry
from madap.echem.procedure import EChemProcedure
from madap.logger import logger

log = logger.get_logger("cyclic_potentiometry")


class Voltammetry_CP(Voltammetry, EChemProcedure):
    def __init__(self, voltage, current, time) -> None:
        super().__init__(voltage, current, time)

    def analyze():
        pass

    def plot():
        # TODO V vs t
        # TODO dQ vs dV
        # TODO v vs Q
        pass

    def save_data():
        pass

    def perform_all_actions():
        pass



from madap.logger import logger
from plotting import Plots


log = logger.get_logger("voltammetry_plotting")

class VoltammetryPlotting(Plots):
    # TODO add corresponding methods to this class
    pass

    # TODO columbic_efficiency vs time
    # TODO Q vs time (capacity(mAh))

    # CP
    # TODO V vs t
    # TODO dQ vs dV
    # TODO v vs Q

    # CA
    # TODO i vs t
    # TODO i vs Q
    # TODO currentDensity vs time
    # TODO Q vs cycle

    #cv
    # TODO plots
    # TODO V vs I


""" This module defines some utility functions for the MADAP project. """
import time
import json
import os

import pandas as pd

from madap.logger import logger



log = logger.get_logger("utils")


def create_dir(directory):
    """Checks if a directory is present, if not creates one at the given location

    Args:
        directory (str): Location where the directory should be created
    """

    if not os.path.exists(directory):
        os.makedirs(directory)
        log.info(f"Created directory {directory}.")
    else:
        log.info(f"Directory {directory} already exists.")
    return directory


def assemble_file_name(*args):
    """Assemble a file name from the given arguments

    Returns:
        str: The assembled file name
    """
    timestamp = time.strftime("%Y%m%d_%H%M%S_")
    return timestamp+"_".join(list(args))

def assemble_data_frame(**kwargs):
    """Assemble a data frame from the given arguments

    Returns:
        Pandas DataFrame: The assembled data frame
    """
    try:
        df = pd.DataFrame.from_dict(kwargs, orient = "index")
        df = df.transpose()
    except AttributeError:
        df = pd.DataFrame(data=kwargs)
    return df

def save_data_as_csv(directory, data, name):
    """Save the given data as csv

    Args:
        directory (str): The directory where the data should be saved
        data (Pandas DataFrame): The data that should be saved
        name (str): The name of the file
    """
    log.info(f"Saving data in {directory} as csv")
    data.to_csv(os.path.join(directory, name))


def save_data_as_json(directory, data, name):
    """Save the given data as json

    Args:
        directory (str): The directory where the data should be saved
        data (dict): The data that should be saved
        name (str): The name of the file
    """
    log.info(f"Saving data in {directory} as json")
    with open(os.path.join(directory, name), 'w', encoding="utf-8") as file:
        json.dump(data, file)
    log.info(f"Saving data in {directory} as json")
    with open(os.path.join(directory, name), 'w', encoding="utf8") as f:
        json.dump(data, f, ensure_ascii=False)

def load_data_as_json(directory, name):
    """ Load the given data as json

    Args:
        directory (str): The directory where the data should be saved
        name (str): The name of the file
    """
    log.info(f"Loading data from {directory} as json")
    with open(os.path.join(directory, name), 'r', encoding="utf-8") as file:
        data = json.load(file)
    return data

def append_to_save_data(directory, added_data, name):
    """Append the given data to the existing data

    Args:
        directory (str): The directory where the data should be saved
        added_data (Pandas DataFrame): The data that should be appended
        name (str): The name of the file
    """
    data = load_data_as_json(directory, name)
    data.update(added_data)
    save_data_as_json(directory, data, name)



"""Impedance Analysis module."""
# for fit the EIS experiments, impedance python package has been used.
# RefMurbach, M., Gerwe, B., Dawson-Elli, N., & Tsui, L. (2020). impedance.py:
# A Python package for electrochemical impedance analysis.
# Journal of Open Source Software, 5(). https://doi.org/10.21105/joss.02349
# cite for EIS fitting: https://github.com/ECSHackWeek/impedance.py
import os
import warnings
import random

import numpy as np
from impedance import validation
from impedance import preprocessing
from impedance.models import circuits
#import impedance.validation as validation
#import impedance.preprocessing as preprocessing
#import impedance.models.circuits as circuits

from attrs import define, field
from attrs.setters import frozen

from madap.logger import logger
from madap.utils import utils
from madap.utils.suggested_circuits import suggested_circuits
from madap.data_acquisition import data_acquisition as da
from madap.echem.procedure import EChemProcedure
from madap.echem.e_impedance.e_impedance_plotting import ImpedancePlotting as iplt

warnings.warn("deprecated", DeprecationWarning)
np.seterr(divide='ignore', invalid='ignore')
# reference the impedance library
log = logger.get_logger("impedance")


@define
class EImpedance:
    """Class for data definition that will be used during the Impedance analysis.
        The data includes the following: frequency, real impedance, imaginary impedance,
        and the phase shift. These attributes are all pandas.Series
        and will stay immutable except the phase shift.
    """
    frequency : list[float] = field(on_setattr=frozen)
    real_impedance : list[float] = field(on_setattr=frozen)
    imaginary_impedance : list[float] = field( on_setattr=frozen)
    phase_shift : list[float] = field(default=None)

    def __repr__(self) -> str:
        """Returns a string representation of the object."""
        return f"Impedance(frequency={self.frequency}, real_impedance={self.real_impedance}, \
                imaginary_impedance={self.imaginary_impedance}, phase_shift={self.phase_shift})"

class EIS(EChemProcedure):
    """General EIS class for the analysis of the EIS data.

    Args:
        EChemProcedure (cls): Parent abstract class
    """
    def __init__(self, impedance, voltage: float = None, suggested_circuit: str = None,
                initial_value = None, max_rc_element: int = 50,
                cut_off: float = 0.85, fit_type: str = 'complex',
                val_low_freq: bool = True, cell_constant="n", max_iterations: int = 5,
                threshold_error:float = 0.009):
        """ Initialize the EIS class.

        Args:
            impedance (np.array): Impedance data containing real and imaginary part.
            voltage (float, optional): Voltage of the EIS measurement. Defaults to None.
            suggested_circuit (str, optional): String defining the suggested circuit. Defaults to None.
            initial_value (list, optional): Initial value of the circuit's element. Defaults to None.
            max_rc_element (int, optional): Maximum number of RC element to be used in the circuit. Defaults to 50.
            cut_off (float, optional): Cut off value of the fitted elements. Defaults to 0.85.
            fit_type (str, optional): Fit type. Defaults to 'complex'.
            val_low_freq (bool, optional): If True, the low frequency is used for the fit. Defaults to True.
            cell_constant (str, optional): Cell constant. Defaults to "n".
            max_iterations (int, optional): Maximum number of iterations for evaluating the accuarcy of fit. Defaults to 5.
        """
        self.impedance = impedance
        self.voltage = voltage
        self.suggested_circuit = suggested_circuit
        self.initial_value = initial_value
        self.max_rc_element = max_rc_element
        self.cut_off = cut_off
        self.fit_type = fit_type
        self.val_low_freq = val_low_freq
        self.cell_constant = cell_constant
        self.max_iterations = max_iterations
        self.threshold_error = threshold_error
        self.conductivity = None
        self.rmse_calc = None
        self.num_rc_linkk = None
        self.eval_fit_linkk = None
        self.z_linkk = None
        self.res_real = None
        self.res_imag = None
        self.chi_val = None
        self.custom_circuit = None
        self.z_fit = None
        self.impedance.phase_shift = self._calculate_phase_shift() if self.impedance.phase_shift is None else self.impedance.phase_shift
        self.figure = None


    # Schönleber, M. et al. A Method for Improving the Robustness of
    # linear Kramers-Kronig Validity Tests.
    # Electrochimica Acta 131, 20–27 (2014) doi: 10.1016/j.electacta.2014.01.034.
    def analyze(self):
        """General function for performing the impedance analysis.
        This will fit the circuit and calculate the conductivity if is applicable.
        """
        f_circuit, z_circuit = np.array(self.impedance.frequency), \
                               np.array(self.impedance.real_impedance +
               1j*self.impedance.imaginary_impedance)

        self.num_rc_linkk, self.eval_fit_linkk , self.z_linkk, \
        self.res_real, self.res_imag = validation.linKK(f_circuit, z_circuit,
                                                        c=self.cut_off,
                                                        max_M=self.max_rc_element,
                                                        fit_type=self.fit_type,
                                                        add_cap=self.val_low_freq)
        self.chi_val = self._chi_calculation()
        log.info(f"Chi value from lin_KK method is {self.chi_val}")

        if any(x < 0 for x in self.impedance.imaginary_impedance):
            f_circuit, z_circuit = preprocessing.ignoreBelowX(f_circuit, z_circuit)

        # if the user did not choose any circuit, some default suggestions will be applied.
        if (self.suggested_circuit and self.initial_value) is None:

            for guess_circuit, guess_value in suggested_circuits.items():
                # apply some random guess
                guess_value = self._initialize_random_guess(guess_value, min(self.impedance.real_impedance))
                # fit the data with random circuit and its randomly guessed elements
                custom_circuit_guess = circuits.CustomCircuit(initial_guess=guess_value, circuit=guess_circuit)

                try:
                    custom_circuit_guess.fit(f_circuit, z_circuit)

                except RuntimeError as exp:
                    log.error(exp)
                    continue

                z_fit_guess = custom_circuit_guess.predict(f_circuit)
                rmse_guess = circuits.fitting.rmse(z_circuit, z_fit_guess)
                log.info(f"With the guessed circuit {guess_circuit} the RMSE error is {rmse_guess}")

                if self.rmse_calc is None:
                    self.rmse_calc = rmse_guess

                if rmse_guess <= self.rmse_calc:
                    self.rmse_calc = rmse_guess
                    self.custom_circuit = custom_circuit_guess
                    self.z_fit = z_fit_guess
        else:
            self.custom_circuit = circuits.CustomCircuit(initial_guess=self.initial_value, circuit=self.suggested_circuit)
            self.custom_circuit.fit(f_circuit, z_circuit)
            self.z_fit = self.custom_circuit.predict(f_circuit)
            self.rmse_calc = circuits.fitting.rmse(z_circuit, self.z_fit)
            log.info(f"With the guessed circuit {self.suggested_circuit} the RMSE error is {self.rmse_calc}")

            # re-evaluating the fit
            iteration = 0
            random_choice = ["add", "subtract"]
            # get the sum of root mean square of the truth values
            rms = np.linalg.norm(z_circuit) / np.sqrt(len(z_circuit))
            while iteration < self.max_iterations:
                iteration += 1
                try:
                    # fit again if the threshold error is not satisfied
                    if self.rmse_calc > (self.threshold_error * rms):
                        # get the initial values according to previous fit and its uncertainty
                        random_selection = random.choice(random_choice)
                        if random_selection == "add":
                            initial_guess_trial = [i+j for i,j in zip(self.custom_circuit.parameters_.tolist(), self.custom_circuit.conf_.tolist())]
                        else:
                            initial_guess_trial = [i-j for i,j in zip(self.custom_circuit.parameters_.tolist(), self.custom_circuit.conf_.tolist())]
                        # refit
                        self.custom_circuit = circuits.CustomCircuit(initial_guess=initial_guess_trial,\
                                                                    circuit=self.suggested_circuit).fit(f_circuit, z_circuit, method="trf")
                        self.z_fit = self.custom_circuit.predict(f_circuit)
                        self.rmse_calc = circuits.fitting.rmse(z_circuit, self.z_fit)
                        log.info(f"With re-evaluating the circuit {self.suggested_circuit} the RMSE error is now {self.rmse_calc}")


                except Exception as e:
                    log.error(e)
                    continue

        if self.cell_constant:
            # calculate the ionic conductivity if cell constant is available
            self.conductivity = self._conductivity_calculation()


    def plot(self, save_dir, plots, optional_name: str = None):
        """Plot the results of the analysis.

        Args:
            save_dir (str): directory where to save the data
            plots (list): list of plot types to be plotted
            optional_name (str, optional): name of the file to be saved. Defaults to None.
        """
        plot_dir = utils.create_dir(os.path.join(save_dir, "plots"))
        plot = iplt()
        fig, available_axes = plot.compose_eis_subplot(plots=plots)

        for sub_ax, plot_name in zip(available_axes, plots):
            if plot_name =="nyquist":
                sub_ax = available_axes[0] if ((len(plots)==4) or (len(plots)==3)) else sub_ax

                plot.nyquist(subplot_ax=sub_ax, frequency=self.impedance.frequency, real_impedance=self.impedance.real_impedance,
                            imaginary_impedance=self.impedance.imaginary_impedance,
                            ax_sci_notation='both', scientific_limit=3, scientific_label_colorbar=False, legend_label=True,
                            voltage=self.voltage, norm_color=True)

            elif plot_name == "nyquist_fit":
                sub_ax = available_axes[2] if len(plots)==4 else \
                        (available_axes[0] if (len(plots)==3 and (("residual" in plots) and ("bode" in plots))) else \
                        (available_axes[1] if (len(plots)==3 and (("residual" in plots) or ("bode" in plots))) else sub_ax))

                plot.nyquist_fit(subplot_ax=sub_ax, frequency=self.impedance.frequency, real_impedance=self.impedance.real_impedance,
                                 imaginary_impedance=self.impedance.imaginary_impedance, fitted_impedance=self.z_fit, chi=self.chi_val,
                                 suggested_circuit=self.custom_circuit.circuit,
                                 ax_sci_notation="both", scientific_limit=3, scientific_label_colorbar=False, legend_label=True,
                                 voltage=self.voltage, norm_color=True)

            elif plot_name == "bode":
                sub_ax = available_axes[1] if len(plots)==4 else (available_axes[1] if (len(plots)==3 and ("residual" in plots)) else
                                                                  (available_axes[2] if (len(plots)==3 and (not "residual" in plots)) else sub_ax))
                plot.bode(subplot_ax=sub_ax, frequency=self.impedance.frequency, real_impedance=self.impedance.real_impedance,
                          imaginary_impedance=self.impedance.imaginary_impedance,
                          phase_shift=self.impedance.phase_shift, ax_sci_notation="y", scientific_limit=3, log_scale="x")

            elif plot_name == "residual":
                sub_ax = available_axes[3] if len(plots)==4 else (available_axes[2] if len(plots)==3 else sub_ax)
                plot.residual(subplot_ax=sub_ax, frequency=self.impedance.frequency, res_real=self.res_real,
                              res_imag=self.res_imag, log_scale='x')

            else:
                log.error("EIS class does not have the selected plot.")
                continue

        fig.tight_layout()
        self.figure = fig
        name = utils.assemble_file_name(optional_name, self.__class__.__name__) if \
                    optional_name else utils.assemble_file_name(self.__class__.__name__)

        plot.save_plot(fig, plot_dir, name)

    def save_data(self, save_dir:str, optional_name:str = None):
        """Save the results of the analysis.

        Args:
            save_dir (str): Directory where the data should be saved.
            optional_name (None): Optional name for the data.
        """
        save_dir = utils.create_dir(os.path.join(save_dir, "data"))
        # Save the fitted circuit
        name = utils.assemble_file_name(optional_name, self.__class__.__name__, "circuit.json") if \
        optional_name else utils.assemble_file_name(self.__class__.__name__, "circuit.json")

        self.custom_circuit.save(os.path.join(save_dir, f"{name}"))
        added_data = {'rc_linKK': self.num_rc_linkk, "eval_fit_linKK": self.eval_fit_linkk, "RMSE_fit_error": self.rmse_calc,
                      "conductivity [S/cm]": self.conductivity, "chi_square": self.chi_val}
        utils.append_to_save_data(directory=save_dir, added_data=added_data, name=name)
        # Save the dataset
        data = utils.assemble_data_frame(**{"frequency [Hz]": self.impedance.frequency,
                                            "impedance [\u03a9]": self.impedance.real_impedance + 1j*self.impedance.imaginary_impedance,
                                            "fit_impedance [\u03a9]": self.z_fit, "residual_real":self.res_real, "residual_imag":self.res_imag,
                                            "Z_linKK [\u03a9]": self.z_linkk})
        data_name = utils.assemble_file_name(optional_name, self.__class__.__name__, "data.csv") if \
                        optional_name else  utils.assemble_file_name(self.__class__.__name__, "data.csv")

        utils.save_data_as_csv(save_dir, data, data_name)

    def perform_all_actions(self, save_dir:str, plots:list, optional_name:str = None):
        """ Wrapper function for executing all action

        Args:
            save_dir (str): Directory where the data should be saved.
            plots (list): List of plot types to be plotted.
        """
        self.analyze()
        self.plot(save_dir, plots, optional_name=optional_name)
        self.save_data(save_dir=save_dir, optional_name=optional_name)

    @property
    def figure(self):
        """Get the figure of the plot.

        Returns:
            obj: Figure object for e_impendance plot.
        """
        return self._figure

    @figure.setter
    def figure(self, figure):
        """Set the figure of the plot.

        Args:
            figure (obj): Figure object for e_impendance plot.
        """
        self._figure = figure

    def _chi_calculation(self):
        """ Calculate the chi value of the fit.

        Returns:
            float: chi square value of the fit
        """
        return np.sum(np.square(self.res_imag) + np.square(self.res_real))

    def _conductivity_calculation(self):
        """ Calculate the conductivity of the circuit.

        Returns:
            float: conductivity of the circuit
        """
        # Check if cell_constant is a float, if not make it a float
        if isinstance(self.cell_constant, str):
            self.cell_constant = float(self.cell_constant)
        conductivity = self.cell_constant * (1/self.custom_circuit.parameters_[0])
        log.info(f"The calculated conductivity is {conductivity} [S.cm⁻¹]")
        return conductivity

    def _calculate_phase_shift(self) -> list[float]:
        """calculate phase shift

        Args:
            imaginary_impedance (class): imaginary impedance data
            real_impedance (class): real impedance data

        Returns:
            phase shift: calculated phase shift based on real and imaginary data
        """

        phase_shift_in_rad = np.arctan(da.format_data(
                             abs(-self.impedance.imaginary_impedance)/da.format_data(abs(self.impedance.real_impedance))))
        return np.rad2deg(phase_shift_in_rad)

    def _initialize_random_guess(self, guess_value, guess_initial_resistance):
        """Initialize the random guess for the circuit's resistace.

        Args:
            guess_value (list): list of initial guessed value without any guess for resistance.
            guess_initial_resistance (float): Value for the initial resistance guess.
        """
        guess_value = [guess_initial_resistance if element == 'x' else element for element in guess_value]
        guess_value = [guess_initial_resistance if element == 'y' else element for element in guess_value]
        guess_value = [guess_initial_resistance if element == 'z' else element for element in guess_value]
        guess_value = [guess_initial_resistance if element == 't' else element for element in guess_value]
        return guess_value

class Mottschotcky(EIS, EChemProcedure):
    """ Class for performing the Mottschotcky procedure.

    Args:
        EIS (class): General EIS class
        EChemProcedure (class): General abstract EChem Procedure class
    """
    def __init__(self, impedance, suggested_circuit: str = None, initial_value=None,
                 max_rc_element: int = 20, cut_off: float = 0.85, fit_type: str = 'complex', val_low_freq=True):
        EIS.__init__(impedance, suggested_circuit, initial_value, max_rc_element, cut_off, fit_type, val_low_freq)

    def analyze(self):
        pass

    def plot(self, save_dir:str, plots:list, optional_name:str):
        pass
    # TODO
    # nyquist
    # fit nyquist
    # bode
    # mott v vs 1/cp_2
    def save_data(self, save_dir:str, optional_name:str):
        pass

    def perform_all_actions(self, save_dir:str, plots:list, optional_name:str):
        pass

class Lissajous(EChemProcedure):
    """ Class for performing the Lissajous procedure."""
    def __init__(self) -> None:
        return None

    def analyze(self):
        pass

    def plot(self, save_dir:str, plots:list, optional_name:str):
        return None
    # TODO
    # legend is the frequency
    # i vs t
    # v vs t
    # i vs E
    def save_data(self, save_dir:str, optional_name:str):
        pass

    def perform_all_actions(self, save_dir:str, plots:list, optional_name:str):
        pass



"""This module is responsible for handaling the data acquisition and data cleaning into MADAP"""
import os
import pandas as pd
import numpy as np

from madap.logger import logger


log = logger.get_logger("data_acquisition")
EXTENSIONS = [".txt", ".csv", ".json", ".xlsx", ".hdf5", ".h5", ".pkl"]

def acquire_data(data_path):
    """Acquire data from a given file

    Args:
        data_path (str): The path to the data

    Returns:
       Pandas DataFrame: Dataframe with extracted data
    """
    _ , extension = os.path.splitext(data_path)

    log.info(f"Importing {extension} file.")

    if extension not in EXTENSIONS:
        log.error(f"Datatype not supported. Supported datatypes are: {EXTENSIONS}")
        raise ValueError("Datatype not supported")

    if extension in (".csv", ".txt"):
        try:
            df = pd.read_csv(data_path, sep=None, engine="python")
        except:
            df = pd.read_csv(data_path, sep=";", engine="python")

    if extension == ".xlsx":
        df = pd.read_excel(data_path)
    if extension == ".json":
        df = pd.read_json(data_path)
    if extension in (".hdf5", ".h5"):
        df = pd.read_hdf(data_path)
    if extension == ".pkl":
        df = pd.read_pickle(data_path)

    return df


def format_data(data):
    """Convert the given data to the array of float

    Args:
        data (_type_): Given data

    Returns:
        A readable format of data for analysis
    """
    if not data is None:
        if isinstance(data, list):
            data = np.array(data, dtype=np.float64)

        if not np.array_equal(data, data.astype(float)):
            data = data.astype(np.float)

    return data


def select_data(data, selected_data:str):
    """ Function to subselect the data from the dataframe

    Args:
        data (DataFrame): The dataframe from which the data is to be selected
        select_data (str): The string containing the start and end row and column

    Returns:
        DataFrame: The subselected dataframe
    """
    numbers = list(map(int, selected_data.split(",")))
    data = data.iloc[:, numbers[2]: numbers[3]].values.reshape(-1)

    # check datatype
    if not isinstance(data, np.ndarray):
        data = np.array(data)
    if isinstance(data[0], str):
        if len(data) == 1:
            data = np.array(eval(data[0]))
        else:
            for i, _ in enumerate(data):
                data[i] = eval(data[i])

    return data


def format_plots(plots):
    """ Format the selection plots into a list

    Args:
        plots (obj): The plots selected by the user

    Returns:
        list: The list of plots selected by the user
    """
    if isinstance(plots, str):
        plots = [plots]
    if isinstance(plots, tuple):
        plots = list(plots)
    return plots

def remove_outlier_specifying_quantile(df, columns, low_quantile = 0.05, high_quantile = 0.95):
    """removing the outliers from the data by specifying the quantile

    Args:
        df (dataframe): original dataframe
        columns (list): colummns for which the outliers are to be removed
        low_quantile (float): lower quantile
        high_quantile (float): upper quantile

    Returns:
        data: the cleaned dataframe
    """
    # select the columns that needs to be studied for outliers
    detect_search = df[columns]
    # Check if the low_quantile and high_quantile are floats, if not convert them to float
    if not isinstance(low_quantile, float):
        low_quantile = float(low_quantile)
    if not isinstance(high_quantile, float):
        high_quantile = float(high_quantile)
    # get the lower quantile of the corresponding columns
    q_low = detect_search.quantile(low_quantile)
    # get the upper quantile of the corresponding columns
    q_high = detect_search.quantile(high_quantile)
    # detect the outliiers in the data and replace those values with nan
    detect_search = detect_search[(detect_search < q_high) & (detect_search > q_low)]
    # find the index of the outliers
    nan_indices = [*set(np.where(np.asanyarray(np.isnan(detect_search)))[0].tolist())]
    log.info(f"The outliers are remove from dataset {detect_search} \
               \n and the indeces are {nan_indices}.")

    return detect_search, nan_indices

def remove_nan_rows(df, nan_indices):
    """ Remove the rows with nan values

    Args:
        df (DataFramo): The dataframe from which the rows are to be removed
        nan_indices (list): The list of indices of the rows to be removed

    Returns:
        DataFrame: The dataframe with the rows removed
    """
    # check if the index of the outliers is present in the dataframe
    available_nan_indices = [i for i in nan_indices if i in df.index.values]
    # removing the rows with nan and reset their indeces
    df = df.drop(df.index[available_nan_indices]).reset_index()
    # delete the columns
    if "index" in df.columns:
        del df["index"]
    remove_unnamed_col(df)
    return df


def remove_unnamed_col(df):
    """ Remove the unnamed columns from the dataframe

    Args:
        df (DataFrame): The dataframe from which unnamed cloums were removed
    """
    if "Unnamed: 0" in df.columns:
        del df["Unnamed: 0"]



"""This module handles the plotting of the Arrhenius procedure"""
from matplotlib import pyplot as plt

from madap.logger import logger
from madap.plotting.plotting import Plots


log = logger.get_logger("arrhenius_plotting")

class ArrheniusPlotting(Plots):
    """General Plotting class for Arrhenius method.

    Args:
        Plots (class): Parent class for plotting all methods.
    """
    def __init__(self) -> None:
        super().__init__()
        self.plot_type = "Arrhenius"

    def arrhenius(self, subplot_ax, temperatures, log_conductivity, inverted_scale_temperatures,
            ax_sci_notation = None, scientific_limit: int = 3):
        """Defines the Arrhenius plot for raw data

        Args:
            subplot_ax (ax): Subplot axis
            temperatures (np.array): Array of temperatures
            log_conductivity (np.array): Array of log conductivity in [S/cm]
            inverted_scale_temperatures (np.array): Array of inverted scale temperatures in [1000/K]
            ax_sci_notation (bool, optional): If True, adds scientific notation to the axis. Defaults to None.
            scientific_limit (int, optional): If ax_sci_notation is True, defines the number of significant digits. Defaults to None.
        """

        log.info("Creating Arrhenius plot")
        subplot_ax2=subplot_ax.twiny()
        self.plot_identity(subplot_ax, xlabel=r"$\frac{1000}{T}$ $[K^{-1}]$",
                           ylabel=r"$\ln$($\sigma$) [$ln(S.cm^{-1})$]",
                           ax_sci_notation=ax_sci_notation,
                           scientific_limit=scientific_limit)

        subplot_ax2.scatter(temperatures, log_conductivity, s=10, c="#20a387ff", rasterized=True)
        self.plot_identity(subplot_ax2, xlabel="$T$ [\u00b0C]")
        self.set_xtick_for_two_axes(subplot_ax, subplot_ax2, [f"{ct:.1f}" for ct in inverted_scale_temperatures], temperatures, invert_axes=True)

    def arrhenius_fit(self, subplot_ax, temperatures, log_conductivity, inverted_scale_temperatures,
                    ln_conductivity_fit, activation, arrhenius_constant, r2_score,
                    ax_sci_notation = None, scientific_limit: int=3):
        """
        Defines the Arrhenius plot for fitted data

        Args:
            subplot_ax (ax): Subplot axis
            temperatures (np.array): Array of temperatures
            log_conductivity (np.array): Array of log conductivity in [S/cm]
            inverted_scale_temperatures (np.array): Array of inverted scale temperatures in [1000/K]
            ln_conductivity_fit (np.array): Array of fitted log conductivity in [S/cm]
            activation (float): Activation energy in [mJ/mol]
            arrhenius_constant (float): Arrhenius constant in [mJ/mol]
            r2_score (float): R2 score of the fit
            ax_sci_notation (bool, optional): Weather or not scientific notation should be adopted for axis. Defaults to None.
            scientific_limit (int, optional): Number of significant digits. Defaults to 3.
        """
        log.info("Creating a fitted Arrhenius plot")
        subplot_ax2=subplot_ax.twiny()
        subplot_ax.plot(inverted_scale_temperatures, ln_conductivity_fit, c="#453781ff", linewidth=1, linestyle="--",
                        label=f" E  = {activation:.2f} [mJ.mol⁻¹] \n A  = {arrhenius_constant:.2f} [S.cm⁻¹] \n R² = {r2_score:.2f}")

        subplot_ax2.scatter(temperatures, log_conductivity, s=10, c="#20a387ff", rasterized=True)
        subplot_ax2.invert_xaxis()
        self.plot_identity(subplot_ax, xlabel=r"$\frac{1000}{T}$ $[K^{-1}]$",
                           ylabel=r"$\ln$($\sigma$) [$ln(S.cm^{-1})$]",
                           ax_sci_notation=ax_sci_notation,
                           scientific_limit=scientific_limit)
        subplot_ax.legend(loc="upper right", fontsize=5.5)

        self.plot_identity(subplot_ax2, xlabel="$T$ [\u00b0C]")

    def compose_arrhenius_subplot(self, plots:list):
        """Creates subplots template for the Arrhenius plots.

        Args:
            plots (list): List of plots to be composed.

        Returns:
            fig, ax: Figure and axis of the subplot.
        """
        plt.close('all')
        if len(plots)==1:
            fig = plt.figure(figsize=(3, 3))
            spec = fig.add_gridspec(1, 1)
            ax = fig.add_subplot(spec[0,0])
            return fig, [ax]

        if len(plots) == 2:
            fig = plt.figure(figsize=(6, 3))
            spec = fig.add_gridspec(1, 2)
            ax1 = fig.add_subplot(spec[0, 0])
            ax2 = fig.add_subplot(spec[0, 1])
            return fig, [ax1, ax2]

        return Exception(f"Number of plots not supported for plot type {self.plot_type}")



"""This module defines the logger for the MADAP application."""
import logging
import os
import sys
APP_LOGGER_NAME = 'MADAP'


def setup_applevel_logger(logger_name = APP_LOGGER_NAME, file_name=None):
    """Sets up the app level logger

    Args:
        logger_name (str, optional): Logger name. Defaults to APP_LOGGER_NAME.
        file_name (str, optional): file name. Defaults to None.

    Returns:
        logger.Logger: app level logger
    """

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(stream_handler)
    if file_name:
        if not os.path.exists("logs"):
            os.makedirs("logs")
        file_handler = logging.FileHandler(f"./logs/{file_name}", encoding="utf-8")
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    return logger

def get_logger(module_name):
    """Returns a logger with the name of the module calling this function

    Args:
        module_name (str): name of the module calling this function

    Returns:
        logging.Logger: logger with the name of the module calling this function
    """
    return logging.getLogger(APP_LOGGER_NAME).getChild(module_name)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst', encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst', encoding="utf-8") as history_file:
    history = history_file.read()

requirements = ["attrs==21.4.0",
        "matplotlib==3.5.3",
        "numpy==1.22.4",
        "pandas==1.4.2",
        "PySimpleGUI==4.60.3",
        "pytest==7.1.2",
        "scikit_learn==1.1.2",
        "impedance==1.4.1"]

test_requirements = ['pytest>=3', ]

setup(
    author="Fuzhan Rahmanian",
    author_email='fuzhanrahmanian@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
    description="This is MADAP, a software package for the analysis of electrochemical data.",
    entry_points={
        'console_scripts': [
            'cli=src.madap_cli:main',
            'gui=src.madap_gui:main',
        ],
    },
    extras_require={
        "dev": ["pytest>=3", ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='madap',
    name='MADAP',
    packages=find_packages(include=['src', 'src.*']),
    #package_dir = {'': 'src'},
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/fuzhanrahmanian/MADAP',
    version='0.9.0',
    zip_safe=False,
)



"""This module implements the GUI application for MADAP """
import io

import PySimpleGUI as sg
import matplotlib.pyplot as plt

from matplotlib.backends.backend_tkagg import FigureCanvasAgg

from madap.utils import gui_elements
from madap_cli import start_procedure
class MadapGui:
    """This class implements the GUI application for MADAP
    """
    eis_plots = ["nyquist" ,"nyquist_fit", "residual", "bode"]
    arrhenius_plots = ["arrhenius", "arrhenius_fit"]

    def __init__(self):
        self.procedure = "Impedance"
        self.impedance_procedure= "EIS"
        self.file = None
        self.results = None
        self.header_list = None
        self.specific= None
        self.plots = None
        self.voltage = None
        self.cell_constant = None
        self.suggested_circuit = None
        self.initial_values = None
        self.upper_limit_quantile = None
        self.lower_limit_quantile = None

    # pylint: disable=inconsistent-return-statements
    def validate_fields(self):
        """ This function validates the fields in the GUI

        Returns:
            bool: Returns false if any of the fields are invalid.
        """
        if self.file == '':
            sg.popup_error('The data path is empty. Select a supported dataset file.', title='Input Error')
            return False
        if self.results == '':
            sg.popup_error('The result path is empty. Select a location for the results.', title='Input Error')
            return False
        if self.plots == []:
            sg.popup_error('Select the desired plot(s).', title='Input Error')
            return False
        if self.procedure == 'Impedance':
            if self.header_list and (len(self.header_list) not in [3,4]):
                sg.popup_error('Wrong number of header inputs.', title='Input Error')
                return False
            if self.specific and (len(self.specific) not in [3,4]):
                sg.popup_error('Wrong number of specific inputs.', title='Input Error')
                return False
        if self.procedure == 'Arrhenius':
            if self.header_list and (len(self.header_list) != 2):
                sg.popup_error('Wrong number of header inputs.', title='Input Error')
                return False
            if self.specific and (len(self.specific) != 2):
                sg.popup_error('Wrong number of specific inputs.', title='Input Error')
                return False
        return True


def draw_figure(element, figure):
    """
    Draws the previously created "figure" in the supplied Image Element

    Args:
        element (PySimpleGUI.Image): The image element to draw the figure on
        figure (matplotlib.figure.Figure): The figure to draw
    """

    plt.close('all')  # erases previously drawn plots
    figure.set_dpi(120)
    canv = FigureCanvasAgg(figure)
    buf = io.BytesIO()
    canv.print_figure(buf, format='png')
    if buf is None:
        return None
    buf.seek(0)
    element.update(data=buf.read())
    return canv


def gui_layout(madap, colors):
    """ This function creates the layout of the GUI

    Args:
        madap (MadapGui): The MadapGui object
        colors (dict): The colors of the GUI

    Returns:
        list: The layout of the layout of the GUI
    """

    # ----------- Create a layout with 3 buttons for the different procedures ----------- #
    layout_buttons = [[ sg.Button("Impedance", key="-BUT_Impedance-", button_color=('white', 'black')),
                        sg.Button("Arrhenius", key="-BUT_Arrhenius-", button_color=colors),
                        sg.Button("Voltammetry", key="-BUT_Voltammetry-", button_color=colors)]]

    # ----------- Create a layout with a field for a data path and a result path ----------- #
    layout_data = [[sg.Text('Data Path', size=(10, 1)), sg.InputText(key='-DATA_PATH-',
                                                                     size=(55,1)),
                    sg.FileBrowse(key='-BROWSE_DATA_PATH-')],
                   [sg.Text('Result Path', size=(10, 1)), sg.InputText(key='-RESULT_PATH-',
                                                                       size=(55,1)),
                    sg.FolderBrowse(key='-BROWSE_RESULT_PATH-')],
                ]

    # ----------- Create a layout with a field for a data selection options ----------- #
    layout_data_selection = [[sg.Text('Headers or specific',justification='left', font=("Arial", 13))],
                             [sg.Combo(['Headers', 'Specific Region'], key='-HEADER_OR_SPECIFIC-',
                                       default_value='Headers')],
                             [sg.InputText(key='-HEADER_OR_SPECIFIC_VALUE-',
                                           tooltip=gui_elements.HEADER_OR_SPECIFIC_HELP,
                                           default_text="freq, real, imag")]]


    # ----------- Create tabs for Impedance procedure ----------- #
    # pylint: disable=unnecessary-comprehension
    tab_layout_eis = [[sg.Text('This are the parameters for the EIS procedure')],
                    [sg.Text('Voltage (optional)',justification='left', font=("Arial", 13),
                             pad=(1,(20,0)))],
                    [ sg.InputText(key="-voltage-", tooltip=gui_elements.VOLTAGE_HELP, enable_events=True),
                     sg.Text('[V]')],
                    [sg.Text('Cell constant (optional)',justification='left', font=("Arial", 13),
                             pad=(1,(20,0)))],
                    [sg.InputText(key="-cell_constant-", tooltip=gui_elements.CELL_CONSTANT_HELP,
                                  enable_events=True), sg.Text('[1/cm]')],
                    [sg.Text("Upper limit of quantile (optional)",justification='left',
                             font=("Arial", 13), pad=(1,(20,0)))],
                    [sg.InputText(key="-upper_limit_quantile-", tooltip=gui_elements.UPPER_LIMIT_QUANTILE_HELP,
                                  enable_events=True, default_text="0.99")],
                    [sg.Text("Lower limit of quantile (optional)",justification='left',
                             font=("Arial", 13), pad=(1,(20,0)))],
                    [sg.InputText(key="-lower_limit_quantile-", tooltip=gui_elements.LOWER_LIMIT_QUANTILE_HELP,
                                  enable_events=True, default_text="0.01")],
                    [sg.Text('Suggeted Circuit',justification='left', font=("Arial", 13),
                             pad=(1,(20,0)))],
                    [sg.InputText(key="-suggested_circuit-", tooltip=gui_elements.SUGGESTED_CIRCUIT_HELP,
                                  default_text="R0-p(R1,CPE1)")],
                    [sg.Text('Initial Value', justification='left', font=("Arial", 13), pad=(1,(20,0)))],
                    [sg.InputText(key="-initial_value-", enable_events=True, tooltip=gui_elements.INITIAL_VALUES_HELP,
                                  default_text="[860, 3e+5, 1e-09, 0.90]")],
                    [sg.Text('Plots',justification='left', font=("Arial", 13), pad=(1,(20,0)))],
                    [sg.Listbox([x for x in madap.eis_plots], key='-PLOTS_Impedance-',
                                size=(50,len(madap.eis_plots)), select_mode=sg.SELECT_MODE_MULTIPLE,
                                expand_x=True, expand_y=True)]]

    tab_layout_liss = [[sg.Text('This is inside Lissajous')],
                [sg.Input(key='-inLiss-')]]

    tab_layout_mott = [[sg.Text('This is inside Mottschosky')],
                    [sg.Input(key='-inMott-')]]

    # ----------- Layout the Impedance Options (Three TABS) ----------- #
    layout_impedance = [[sg.TabGroup(
                        [[sg.Tab('EIS', tab_layout_eis, key='-TAB_EIS-', expand_y=True),
                        sg.Tab('Lissajous', tab_layout_liss,  background_color='darkred',
                               key='-TAB_Lissajous-', expand_y=True),
                        sg.Tab('Mottschosky', tab_layout_mott, background_color='darkgreen',
                               key='-TAB_Mottschosky-', expand_y=True)]],
                        tab_location='topleft', selected_title_color='black', enable_events=True,
                        expand_y=True)]]

    # ----------- Layout the Arrhenius Options ----------- #
    layout_arrhenius = [[sg.Text('This are the parameters for the Arrhenius procedure')],
                        [sg.Text('Plots',justification='left', font=("Arial", 13), pad=(1,(20,0)))],
                        [sg.Listbox([x for x in madap.arrhenius_plots], key='-PLOTS_Arrhenius-',
                                    size=(50,len(madap.arrhenius_plots)+1),
                                    select_mode=sg.SELECT_MODE_MULTIPLE, expand_x=True,
                                    expand_y=True)]]

    # ----------- TODO Layout the Voltammetry Options ----------- #
    layout_voltammetry = [[sg.Text('This is Voltammetry')]]


    # ----------- Assemble the Procedure Column Element with the three layouts ----------- #
    procedure_column = [[sg.Column(layout_impedance, key='-COL_Impedance-', scrollable=True,
                                   vertical_scroll_only=True, expand_x=True, expand_y=True),
                        sg.Column(layout_arrhenius, visible=False, key='-COL_Arrhenius-',
                                  scrollable=True, vertical_scroll_only=True, expand_x=True,
                                  expand_y=True),
                        sg.Column(layout_voltammetry, visible=False, key='-COL_Voltammetry-',
                                  scrollable=True, vertical_scroll_only=True, expand_x=True,
                                  expand_y=True)]]

    # ----------- Assemble the left Column Element ----------- #
    col1 = sg.Column([[sg.Frame('Data Selection:', layout_data_selection, font=("Arial", 15),
                                size=(550, 120), expand_y=True)],
                      [sg.Frame('Methods:', procedure_column, font=("Arial", 15), size=(550, 630),
                                expand_y=True)]],
                      expand_x=True, expand_y=True)

    # ----------- Layout the right Column Element ----------- #
    col2 = sg.Column([[sg.Frame('Plots:', [[sg.Image(key='-IMAGE-')]], visible=False,
                                key='-COL_PLOTS-')]])

    # ----------- Assemble the main layout ----------- #
    layout = [
        [layout_buttons],
        [layout_data],
        [col1, col2],
        [sg.Text('',justification='left', font=("Arial", 13), pad=(1,(20,0)), key='-LOG-',
                 enable_events=True)],
        [sg.Button('RUN'), sg.Button('EXIT')]]

    return layout


def main():
    """Main function of the GUI
    """

    # Select a theme
    sg.theme("LightGreen6")

    # Create class with initial values
    madap_gui = MadapGui()

    # Get primary colors and assemble window
    colors = (sg.theme_text_color(), sg.theme_background_color())
    layout = gui_layout(madap_gui, colors)
    title = 'MADAP: Modular Automatic Data Analysis Platform'
    window = sg.Window(title, layout, resizable=True)
    # Event loop
    while True:
        event, values = window.read()
        print(event, values)
        if event in (sg.WIN_CLOSED, 'EXIT'):
            break
        if event in ['-BUT_Impedance-', '-BUT_Arrhenius-', '-BUT_Voltammetry-']:
            event = event.strip('-BUT_')
            window[f'-COL_{madap_gui.procedure}-'].update(visible=False)
            window[f'-BUT_{madap_gui.procedure}-'].update(button_color=colors)
            window[f'-COL_{event}-'].update(visible=True)
            window[f'-BUT_{event}-'].update(button_color=('white', 'black'))
            madap_gui.procedure = event
        if values[0] in ['-TAB_EIS-', '-TAB_Lissajous-', '-TAB_Mottschotcky-']:
            madap_gui.impedance_procedure = values[0].strip('-TAB_')

        # Prevent the user from inoutting a value that is not a number in the voltage, cell constant and initial_value input field
        if event == '-voltage-' and len(values['-voltage-']) \
                                and values['-voltage-'][-1] not in '012345678890,.':
            window['-voltage-'].update(values['-voltage-'][:-1])
        if event == '-cell_constant-' and len(values['-cell_constant-']) \
                                      and values['-cell_constant-'][-1] not in '012345678890,.':
            window['-cell_constant-'].update(values['-cell_constant-'][:-1])
        if event == '-upper_limit_quantile-' and len(values['-upper_limit_quantile-']) \
                                             and values['-upper_limit_quantile-'][-1] not in '012345678890,.':
            window['-upper_limit_quantile-'].update(values['-upper_limit_quantile-'][:-1])
        if event == '-lower_limit_quantile-' and len(values['-lower_limit_quantile-']) \
                                             and values['-lower_limit_quantile-'][-1] not in '012345678890,.':
            window['-lower_limit_quantile-'].update(values['-lower_limit_quantile-'][:-1])

        if event == '-initial_value-' and len(values['-initial_value-']) \
                                      and values['-initial_value-'][-1] not in '012345678890,.e-+[]':
            window['-initial_value-'].update(values['-initial_value-'][:-1])
        if event == 'RUN':
            window['-LOG-'].update('Starting procedure...')
            madap_gui.file = values['-DATA_PATH-']
            madap_gui.results = values['-RESULT_PATH-']
            madap_gui.plots = values[f'-PLOTS_{madap_gui.procedure}-']
            madap_gui.voltage = values['-voltage-']
            madap_gui.cell_constant = values['-cell_constant-']
            madap_gui.suggested_circuit = values['-suggested_circuit-'] \
                                          if not values['-suggested_circuit-'] == '' else None
            madap_gui.initial_values = values['-initial_value-'] \
                                          if not values['-initial_value-'] == '' else None
            madap_gui.upper_limit_quantile = values['-upper_limit_quantile-'] \
                                          if not values['-upper_limit_quantile-'] == '' else None
            madap_gui.lower_limit_quantile = values['-lower_limit_quantile-'] \
                                          if not values['-lower_limit_quantile-'] == '' else None

            if values['-HEADER_OR_SPECIFIC-'] == 'Headers':
                madap_gui.specific = None
                madap_gui.header_list = values['-HEADER_OR_SPECIFIC_VALUE-'].replace(" ","")
                madap_gui.header_list = list(madap_gui.header_list.split(','))
            else:
                madap_gui.header_list = None
                madap_gui.specific = values['-HEADER_OR_SPECIFIC_VALUE-'].replace(" ","")
                madap_gui.specific = list(madap_gui.specific.split(';'))

            # Validate the fields
            validation = madap_gui.validate_fields()
            if not validation:
                window['-LOG-'].update('Inputs were not valid! Try again.')
                continue
            try:
                procedure = start_procedure(madap_gui)
            except Exception as e:
                sg.popup(f'Error: Something went wrong. {e}')
                continue
            window['-LOG-'].update('Generating plot...')
            window['-COL_PLOTS-'].update(visible=True)
            window['-IMAGE-']('')
            draw_figure(window['-IMAGE-'], procedure.figure)
            window['-LOG-'].update('DONE! Results and plots were saved in the given path')


if __name__ == '__main__':
    main()
